# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::Events;

use strict;

require Exporter;
@Katie::Events::ISA = qw(Exporter);
@Katie::Events::EXPORT = qw(
  create_event
  append_event_comment
  destroy_event
);

# ensure the comment is blank or ends with \n
sub fix_comment {
  my($comment) = @_;

  $$comment .= "\n" unless $$comment eq '' || $$comment =~ /\n$/s;
}

sub create_event {
  my($vob_db, $object_id, $event_type, $uid, $comment) = @_;

  fix_comment(\$comment);

  $vob_db->insert("event", {
    object_id  => $object_id,
    event_type => $event_type,
    uid        => $uid,
    comment    => $comment,
  });
  return $vob_db->get_seq_currval("event_id_seq");
}

sub append_event_comment {
  my($vob_db, $event_id, $comment) = @_;

  fix_comment(\$comment);

  $comment = $vob_db->escape_sql($comment);
  $vob_db->exec("
      UPDATE event
      SET    comment = comment || '$comment'
      WHERE  id = $event_id;
  ");
}

sub destroy_event {
  my($vob_db, $event_id) = @_;

  $vob_db->exec("DELETE FROM event WHERE id = $event_id;");
}

1;
