# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::SunRPCServer;

use strict;

use FileHandle;
use Carp;
use File::stat;
use Katie::Constants;

use Inline C => 'DATA',
           NAME => 'Katie::SunRPCServer',
           INC => '-I' . PREFIX_DIR . '/include/katie',
           LIBS => '-lkatiesunrpcserver',
           DIRECTORY => PREFIX_DIR . '/lib/katie';

sub new {
  shift(@_);
  my(%args) = @_;
  my($self) = {};

  $self->{fd} = init(
    $args{mnt},
    $args{getattr},
    $args{setattr},
    $args{lookup},
    $args{read},
    $args{write},
    $args{create},
    $args{remove},
    $args{rename},
    $args{link},
    $args{mkdir},
    $args{rmdir},
    $args{readdir},
    $args{statfs},
  );

  bless $self;
  return $self;
}

sub fd {
  my($self) = @_;

  return $self->{fd};
}

sub handle {
  my($self, $readfds) = @_;

  handle_c($readfds);
}

1;

__DATA__

__C__

#include "katie_nfs.h"
#include "katie_mount.h"
#include <rpc/pmap_clnt.h>

static SV *mnt_handler = (SV*)NULL;
static SV *getattr_handler = (SV*)NULL;
static SV *setattr_handler = (SV*)NULL;
static SV *lookup_handler = (SV*)NULL;
static SV *read_handler = (SV*)NULL;
static SV *write_handler = (SV*)NULL;
static SV *create_handler = (SV*)NULL;
static SV *remove_handler = (SV*)NULL;
static SV *rename_handler = (SV*)NULL;
static SV *link_handler = (SV*)NULL;
static SV *mkdir_handler = (SV*)NULL;
static SV *rmdir_handler = (SV*)NULL;
static SV *readdir_handler = (SV*)NULL;
static SV *statfs_handler = (SV*)NULL;

static int uid_hash;
static int gid_hash;
static int groups_hash;

/* map from NFS file types to Unix modes */
u_int ftype_map[9] =
{
  0 << 12, /* NFNON */
  8 << 12, /* NFREG */
  4 << 12, /* NFDIR */
  6 << 12, /* NFBLK */
  2 << 12, /* NFCHR */
 10 << 12, /* NFLNK */
 12 << 12, /* NFSOCK */
  0 << 12, /* NFBAD */
  1 << 12, /* NFFIFO */
};

void *mountproc_null_1_svc(void *argp, struct svc_req *rqstp)
{
  static char result = 0;
  return (void *) &result;
}

fhstatus *mountproc_mnt_1_svc(dirpath *argp, struct svc_req *rqstp)
{
  char argbuf[MNTPATHLEN + 1];
  static fhstatus result;
  char *fhandle;
  STRLEN len;

  dSP;
  int count;
  SV  *sv;

  strncpy(argbuf, *argp, MNTPATHLEN);
  argbuf[MNTPATHLEN] = '\0';

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(sv_2mortal(newSVpv(argbuf, 0)));
  PUTBACK;

  count = perl_call_sv(mnt_handler, G_SCALAR);

  SPAGAIN;

  if (count != 1) {
    croak("mnt returned %d arguments\n", count);
  }

  sv = POPs;
  fhandle = SvPV(sv, len);
  if (len == 32) {
    result.fhs_status = 0;
    memcpy(result.fhstatus_u.fhs_fhandle, fhandle, 32);
  } else {
    result.fhs_status = SvIV(sv);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

mountlist *mountproc_dump_1_svc(void *argp, struct svc_req *rqstp)
{
  printf("mountproc_dump_1_svc called\n");
  exit(1);
}

void *mountproc_umnt_1_svc(dirpath *argp, struct svc_req *rqstp)
{
  static char result = 0;
  return (void *) &result;
}

void *mountproc_umntall_1_svc(void *argp, struct svc_req *rqstp)
{
  printf("mountproc_umntall_1_svc called\n");
  exit(1);
}

exports *mountproc_export_1_svc(void *argp, struct svc_req *rqstp)
{
  printf("mountproc_export_1_svc called\n");
  exit(1);
}

exports *mountproc_exportall_1_svc(void *argp, struct svc_req *rqstp)
{
  printf("mountproc_exportall_1_svc called\n");
  exit(1);
}

ppathcnf *mountproc_pathconf_2_svc(dirpath *argp, struct svc_req *rqstp)
{
  printf("mountproc_pathconf_2_svc called\n");
  exit(1);
}

void *nfsproc_null_2_svc(void *argp, struct svc_req *rqstp)
{
  static char result = 0;
  return (void *) &result;
}

SV *make_auth_hash_ref(struct authunix_parms *unix_cred)
{
  SV *uid;
  SV *gid;
  SV *groups_array[NGRPS];
  AV *groups_AV;
  SV *groups;
  HV *auth;
  int count;

  /* get the uid, gid, and groups into the appropriate SV's */
  uid = newSViv(unix_cred->aup_uid);
  gid = newSViv(unix_cred->aup_gid);
  for (count = 0; count < unix_cred->aup_len; count++) {
    groups_array[count] = sv_2mortal(newSViv(unix_cred->aup_gids[count]));
  }

  /* create the groups AV and then the reference to the AV */
  groups_AV = (AV*)sv_2mortal((SV*)av_make(unix_cred->aup_len, groups_array));
  groups = newRV_inc((SV*)groups_AV);

  /* put uid, gid, and groups into the auth hash */
  auth = (HV*)sv_2mortal((SV*)newHV());
  hv_store(auth, "uid", 3, uid, uid_hash);
  hv_store(auth, "gid", 3, gid, gid_hash);
  hv_store(auth, "groups", 6, groups, groups_hash);

  /* return a reference to the auth hash */
  return sv_2mortal(newRV_inc((SV*)auth));
}

attrstat *nfsproc_getattr_2_svc(nfs_fh *argp, struct svc_req *rqstp)
{
  static attrstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->data), NFS_FHSIZE)));
  PUTBACK;

  count = perl_call_sv(getattr_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("result.status == 0\n");
    }
  } else if (count == 14) {
    result.status = 0;
    result.attrstat_u.attributes.ctime.seconds = POPi;
    result.attrstat_u.attributes.ctime.useconds = 0;
    result.attrstat_u.attributes.mtime.seconds = POPi;
    result.attrstat_u.attributes.mtime.useconds = 0;
    result.attrstat_u.attributes.atime.seconds = POPi;
    result.attrstat_u.attributes.atime.useconds = 0;
    result.attrstat_u.attributes.fileid = POPi;
    result.attrstat_u.attributes.fsid = POPi;
    result.attrstat_u.attributes.blocks = POPi;
    result.attrstat_u.attributes.rdev = POPi;
    result.attrstat_u.attributes.blocksize = POPi;
    result.attrstat_u.attributes.size = POPi;
    result.attrstat_u.attributes.gid = POPi;
    result.attrstat_u.attributes.uid = POPi;
    result.attrstat_u.attributes.nlink = POPi;
    result.attrstat_u.attributes.mode = POPi;
    result.attrstat_u.attributes.type = POPi;
    result.attrstat_u.attributes.mode |=
        ftype_map[result.attrstat_u.attributes.type];
  } else {
    croak("getattr count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

attrstat *nfsproc_setattr_2_svc(sattrargs *argp, struct svc_req *rqstp)
{
  static attrstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->file.data), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.mode)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.uid)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.gid)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.size)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.atime.seconds)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.mtime.seconds)));
  PUTBACK;

  count = perl_call_sv(setattr_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("result.status == 0\n");
    }
  } else if (count == 14) {
    result.status = 0;
    result.attrstat_u.attributes.ctime.seconds = POPi;
    result.attrstat_u.attributes.ctime.useconds = 0;
    result.attrstat_u.attributes.mtime.seconds = POPi;
    result.attrstat_u.attributes.mtime.useconds = 0;
    result.attrstat_u.attributes.atime.seconds = POPi;
    result.attrstat_u.attributes.atime.useconds = 0;
    result.attrstat_u.attributes.fileid = POPi;
    result.attrstat_u.attributes.fsid = POPi;
    result.attrstat_u.attributes.blocks = POPi;
    result.attrstat_u.attributes.rdev = POPi;
    result.attrstat_u.attributes.blocksize = POPi;
    result.attrstat_u.attributes.size = POPi;
    result.attrstat_u.attributes.gid = POPi;
    result.attrstat_u.attributes.uid = POPi;
    result.attrstat_u.attributes.nlink = POPi;
    result.attrstat_u.attributes.mode = POPi;
    result.attrstat_u.attributes.type = POPi;
    result.attrstat_u.attributes.mode |=
        ftype_map[result.attrstat_u.attributes.type];
  } else {
    croak("setattr count is %d\n", count);
  }
  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

void *nfsproc_root_2_svc(void *argp, struct svc_req *rqstp)
{
  printf("nfsproc_root_2_svc called\n");
  exit(1);
}

diropres *nfsproc_lookup_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  static diropres result;
  char *fhandle;
  STRLEN len;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;
  SV *sv;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->name, 0)));
  PUTBACK;

  count = perl_call_sv(lookup_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("nfs_lookup result.status == 0\n");
    }
  } else if (count == 15) {
    result.status = 0;
    result.diropres_u.diropres.attributes.ctime.seconds = POPi;
    result.diropres_u.diropres.attributes.ctime.useconds = 0;
    result.diropres_u.diropres.attributes.mtime.seconds = POPi;
    result.diropres_u.diropres.attributes.mtime.useconds = 0;
    result.diropres_u.diropres.attributes.atime.seconds = POPi;
    result.diropres_u.diropres.attributes.atime.useconds = 0;
    result.diropres_u.diropres.attributes.fileid = POPi;
    result.diropres_u.diropres.attributes.fsid = POPi;
    result.diropres_u.diropres.attributes.blocks = POPi;
    result.diropres_u.diropres.attributes.rdev = POPi;
    result.diropres_u.diropres.attributes.blocksize = POPi;
    result.diropres_u.diropres.attributes.size = POPi;
    result.diropres_u.diropres.attributes.gid = POPi;
    result.diropres_u.diropres.attributes.uid = POPi;
    result.diropres_u.diropres.attributes.nlink = POPi;
    result.diropres_u.diropres.attributes.mode = POPi;
    result.diropres_u.diropres.attributes.type = POPi;
    result.diropres_u.diropres.attributes.mode |=
        ftype_map[result.diropres_u.diropres.attributes.type];
    sv = POPs;
    fhandle = SvPV(sv, len);
    if (len != 32) {
      croak("nfs_lookup fhandle len == %d\n", len);
    }
    memcpy(result.diropres_u.diropres.file.data, fhandle, 32);
  } else {
    croak("nfs_lookup count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

readlinkres *nfsproc_readlink_2_svc(nfs_fh *argp, struct svc_req *rqstp)
{
  printf("nfsproc_readlink_2_svc called\n");
  exit(1);
}

readres *nfsproc_read_2_svc(readargs *argp, struct svc_req *rqstp)
{
  static readres result;
  static char result_data[NFS_MAXDATA];
  char *data;
  STRLEN len;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;
  SV *sv;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->file), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSViv(argp->offset)));
  XPUSHs(sv_2mortal(newSViv(argp->count > NFS_MAXDATA
                            ? NFS_MAXDATA
                            : argp->count)));
  PUTBACK;

  count = perl_call_sv(read_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("nfs_read result.status == 0\n");
    }
  } else if (count == 15) {
    result.status = 0;
    result.readres_u.reply.attributes.ctime.seconds = POPi;
    result.readres_u.reply.attributes.ctime.useconds = 0;
    result.readres_u.reply.attributes.mtime.seconds = POPi;
    result.readres_u.reply.attributes.mtime.useconds = 0;
    result.readres_u.reply.attributes.atime.seconds = POPi;
    result.readres_u.reply.attributes.atime.useconds = 0;
    result.readres_u.reply.attributes.fileid = POPi;
    result.readres_u.reply.attributes.fsid = POPi;
    result.readres_u.reply.attributes.blocks = POPi;
    result.readres_u.reply.attributes.rdev = POPi;
    result.readres_u.reply.attributes.blocksize = POPi;
    result.readres_u.reply.attributes.size = POPi;
    result.readres_u.reply.attributes.gid = POPi;
    result.readres_u.reply.attributes.uid = POPi;
    result.readres_u.reply.attributes.nlink = POPi;
    result.readres_u.reply.attributes.mode = POPi;
    result.readres_u.reply.attributes.type = POPi;
    result.readres_u.reply.attributes.mode |=
        ftype_map[result.readres_u.reply.attributes.type];
    sv = POPs;
    data = SvPV(sv, len);
    memcpy(result_data, data, len);
    result.readres_u.reply.data.data_val = result_data;
    result.readres_u.reply.data.data_len = len;
  } else {
    croak("nfs_read count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

void *nfsproc_writecache_2_svc(void *argp, struct svc_req *rqstp)
{
  printf("nfsproc_writecache_2_svc called\n");
  exit(1);
}

attrstat *nfsproc_write_2_svc(writeargs *argp, struct svc_req *rqstp)
{
  static attrstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->file), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSViv(argp->offset)));
  XPUSHs(sv_2mortal(newSVpv(argp->data.data_val, argp->data.data_len)));
  PUTBACK;

  count = perl_call_sv(write_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("nfs_write result.status == 0\n");
    }
  } else if (count == 14) {
    result.status = 0;
    result.attrstat_u.attributes.ctime.seconds = POPi;
    result.attrstat_u.attributes.ctime.useconds = 0;
    result.attrstat_u.attributes.mtime.seconds = POPi;
    result.attrstat_u.attributes.mtime.useconds = 0;
    result.attrstat_u.attributes.atime.seconds = POPi;
    result.attrstat_u.attributes.atime.useconds = 0;
    result.attrstat_u.attributes.fileid = POPi;
    result.attrstat_u.attributes.fsid = POPi;
    result.attrstat_u.attributes.blocks = POPi;
    result.attrstat_u.attributes.rdev = POPi;
    result.attrstat_u.attributes.blocksize = POPi;
    result.attrstat_u.attributes.size = POPi;
    result.attrstat_u.attributes.gid = POPi;
    result.attrstat_u.attributes.uid = POPi;
    result.attrstat_u.attributes.nlink = POPi;
    result.attrstat_u.attributes.mode = POPi;
    result.attrstat_u.attributes.type = POPi;
    result.attrstat_u.attributes.mode |=
        ftype_map[result.attrstat_u.attributes.type];
  } else {
    croak("nfs_write count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

diropres *nfsproc_create_2_svc(createargs *argp, struct svc_req *rqstp)
{
  static diropres result;
  char *fhandle;
  STRLEN len;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;
  SV *sv;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->where.dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->where.name, 0)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.mode)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.uid)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.gid)));
  /* size, atime, mtime are ignored */
  PUTBACK;

  count = perl_call_sv(create_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("nfs_create result.status == 0\n");
    }
  } else if (count == 15) {
    result.status = 0;
    result.diropres_u.diropres.attributes.ctime.seconds = POPi;
    result.diropres_u.diropres.attributes.ctime.useconds = 0;
    result.diropres_u.diropres.attributes.mtime.seconds = POPi;
    result.diropres_u.diropres.attributes.mtime.useconds = 0;
    result.diropres_u.diropres.attributes.atime.seconds = POPi;
    result.diropres_u.diropres.attributes.atime.useconds = 0;
    result.diropres_u.diropres.attributes.fileid = POPi;
    result.diropres_u.diropres.attributes.fsid = POPi;
    result.diropres_u.diropres.attributes.blocks = POPi;
    result.diropres_u.diropres.attributes.rdev = POPi;
    result.diropres_u.diropres.attributes.blocksize = POPi;
    result.diropres_u.diropres.attributes.size = POPi;
    result.diropres_u.diropres.attributes.gid = POPi;
    result.diropres_u.diropres.attributes.uid = POPi;
    result.diropres_u.diropres.attributes.nlink = POPi;
    result.diropres_u.diropres.attributes.mode = POPi;
    result.diropres_u.diropres.attributes.type = POPi;
    result.diropres_u.diropres.attributes.mode |=
        ftype_map[result.diropres_u.diropres.attributes.type];
    sv = POPs;
    fhandle = SvPV(sv, len);
    if (len != 32) {
      croak("nfs_create fhandle len == %d\n", len);
    }
    memcpy(result.diropres_u.diropres.file.data, fhandle, 32);
  } else {
    croak("nfs_create count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

nfsstat *nfsproc_remove_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  static nfsstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->name, 0)));
  PUTBACK;

  count = perl_call_sv(remove_handler, G_SCALAR);

  SPAGAIN;

  if (count != 1) {
    croak("nfs_remove count is %d\n", count);
  }

  result = POPi;

  return &result;
}

nfsstat *nfsproc_rename_2_svc(renameargs *argp, struct svc_req *rqstp)
{
  static nfsstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->from.dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->from.name, 0)));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->to.dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->to.name, 0)));
  PUTBACK;

  count = perl_call_sv(rename_handler, G_ARRAY);

  SPAGAIN;

  if (count != 1) {
    croak("nfs_rename count is %d\n", count);
  }

  result = POPi;

  return &result;
}

nfsstat *nfsproc_link_2_svc(linkargs *argp, struct svc_req *rqstp)
{
  static nfsstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->from), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->to.dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->to.name, 0)));
  PUTBACK;

  count = perl_call_sv(link_handler, G_ARRAY);

  SPAGAIN;

  if (count != 1) {
    croak("nfs_link count is %d\n", count);
  }

  result = POPi;

  return &result;
}

nfsstat *nfsproc_symlink_2_svc(symlinkargs *argp, struct svc_req *rqstp)
{
  printf("nfsproc_symlink_2_svc called\n");
  exit(1);
}

/* except for the Perl call, this is the same as nfsproc_create_2_svc */
diropres *nfsproc_mkdir_2_svc(createargs *argp, struct svc_req *rqstp)
{
  static diropres result;
  char *fhandle;
  STRLEN len;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;
  SV *sv;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->where.dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->where.name, 0)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.mode)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.uid)));
  XPUSHs(sv_2mortal(newSViv(argp->attributes.gid)));
  /* size, atime, mtime are ignored */
  PUTBACK;

  count = perl_call_sv(mkdir_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("nfs_create result.status == 0\n");
    }
  } else if (count == 15) {
    result.status = 0;
    result.diropres_u.diropres.attributes.ctime.seconds = POPi;
    result.diropres_u.diropres.attributes.ctime.useconds = 0;
    result.diropres_u.diropres.attributes.mtime.seconds = POPi;
    result.diropres_u.diropres.attributes.mtime.useconds = 0;
    result.diropres_u.diropres.attributes.atime.seconds = POPi;
    result.diropres_u.diropres.attributes.atime.useconds = 0;
    result.diropres_u.diropres.attributes.fileid = POPi;
    result.diropres_u.diropres.attributes.fsid = POPi;
    result.diropres_u.diropres.attributes.blocks = POPi;
    result.diropres_u.diropres.attributes.rdev = POPi;
    result.diropres_u.diropres.attributes.blocksize = POPi;
    result.diropres_u.diropres.attributes.size = POPi;
    result.diropres_u.diropres.attributes.gid = POPi;
    result.diropres_u.diropres.attributes.uid = POPi;
    result.diropres_u.diropres.attributes.nlink = POPi;
    result.diropres_u.diropres.attributes.mode = POPi;
    result.diropres_u.diropres.attributes.type = POPi;
    result.diropres_u.diropres.attributes.mode |=
        ftype_map[result.diropres_u.diropres.attributes.type];
    sv = POPs;
    fhandle = SvPV(sv, len);
    if (len != 32) {
      croak("nfs_create fhandle len == %d\n", len);
    }
    memcpy(result.diropres_u.diropres.file.data, fhandle, 32);
  } else {
    croak("nfs_create count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

nfsstat *nfsproc_rmdir_2_svc(diropargs *argp, struct svc_req *rqstp)
{
  static nfsstat result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSVpv(argp->name, 0)));
  PUTBACK;

  count = perl_call_sv(rmdir_handler, G_SCALAR);

  SPAGAIN;

  if (count != 1) {
    croak("nfs_rmdir count is %d\n", count);
  }

  result = POPi;

  return &result;
}

readdirres *nfsproc_readdir_2_svc(readdirargs *argp, struct svc_req *rqstp)
{
  static readdirres result;
  entry  **ep, *e;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int   count;
  SV*   sv;
  char  *s;
  int   len;
  u_int cookie;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  xdr_free((xdrproc_t) xdr_readdirres, (caddr_t) &result);

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->dir), NFS_FHSIZE)));
  XPUSHs(sv_2mortal(newSViv(*(uint *)argp->cookie)));
  XPUSHs(sv_2mortal(newSViv(argp->count / (sizeof(entry) + 16))));
  PUTBACK;

  count = perl_call_sv(readdir_handler, G_ARRAY);

  SPAGAIN;

  if (count == 0) {
    result.status = 0;
    result.readdirres_u.reply.entries = NULL;
    result.readdirres_u.reply.eof = 1;
  } else if (count == 1) {
    result.status = POPi;
    if (result.status == 0) {
      croak("result.status == 0\n");
    }
  } else if (count % 3 == 1) {
    ep = &(result.readdirres_u.reply.entries);
    while (count > 1) {
      if (!(e = *ep = (entry *)malloc(sizeof(entry)))) {
        fprintf(stderr, "malloc failed\n");
        exit(1);
      }

      /* cookie */
      cookie = POPi;
      memcpy(&e->cookie, &cookie, sizeof(nfscookie));

      /* name */
      sv = POPs;
      s = SvPV(sv, len);
      if (!(e->name = malloc(len + 1))) {
        fprintf(stderr, "malloc failed\n");
      }
      strcpy(e->name, s);

      /* fileid */
      e->fileid = POPi;

      count -= 3;
      ep = &e->nextentry;
    }
    *ep = NULL;
    result.readdirres_u.reply.eof = POPi;
  } else {
    croak("readdir count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

statfsres *nfsproc_statfs_2_svc(nfs_fh *argp, struct svc_req *rqstp)
{
  static statfsres result;
  struct authunix_parms *unix_cred =
      (struct authunix_parms *) rqstp->rq_clntcred;

  dSP;
  int count;

  if (rqstp->rq_cred.oa_flavor != AUTH_UNIX) {
    fprintf(stderr, "unable to handle auth != AUTH_UNIX\n");
    exit(1);
  }

  ENTER;
  SAVETMPS;

  PUSHMARK(SP);
  XPUSHs(make_auth_hash_ref(unix_cred));
  XPUSHs(sv_2mortal(newSVpv((char *)&(argp->data), NFS_FHSIZE)));
  PUTBACK;

  count = perl_call_sv(statfs_handler, G_ARRAY);

  SPAGAIN;

  if (count == 1) {
    if (!(result.status = POPi)) {
      croak("result.status == 0\n");
    }
  } else if (count == 5) {
    result.status = 0;
    result.statfsres_u.reply.bavail = POPi;
    result.statfsres_u.reply.bfree = POPi;
    result.statfsres_u.reply.blocks = POPi;
    result.statfsres_u.reply.bsize = POPi;
    result.statfsres_u.reply.tsize = POPi;
  } else {
    croak("statfs count is %d\n", count);
  }

  PUTBACK;
  FREETMPS;
  LEAVE;

  return &result;
}

void dispatch_mount(struct svc_req *rqstp, register SVCXPRT *transp)
{
  union {
    dirpath mountproc_mnt_1_arg;
    dirpath mountproc_umnt_1_arg;
  } argument;
  char *result;
  xdrproc_t _xdr_argument, _xdr_result;
  char *(*local)(char *, struct svc_req *);

  switch (rqstp->rq_proc) {

    case MOUNTPROC_NULL:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) mountproc_null_1_svc;
      break;

    case MOUNTPROC_MNT:
      _xdr_argument = (xdrproc_t) xdr_dirpath;
      _xdr_result = (xdrproc_t) xdr_fhstatus;
      local = (char *(*)(char *, struct svc_req *)) mountproc_mnt_1_svc;
      break;

    case MOUNTPROC_DUMP:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_mountlist;
      local = (char *(*)(char *, struct svc_req *)) mountproc_dump_1_svc;
      break;

    case MOUNTPROC_UMNT:
      _xdr_argument = (xdrproc_t) xdr_dirpath;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) mountproc_umnt_1_svc;
      break;

    case MOUNTPROC_UMNTALL:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) mountproc_umntall_1_svc;
      break;

    case MOUNTPROC_EXPORT:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_exports;
      local = (char *(*)(char *, struct svc_req *)) mountproc_export_1_svc;
      break;

    case MOUNTPROC_EXPORTALL:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_exports;
      local = (char *(*)(char *, struct svc_req *)) mountproc_exportall_1_svc;
      break;

    case MOUNTPROC_PATHCONF:
      _xdr_argument = (xdrproc_t) xdr_dirpath;
      _xdr_result = (xdrproc_t) xdr_ppathcnf;
      local = (char *(*)(char *, struct svc_req *)) mountproc_pathconf_2_svc;
      break;

    default:
      svcerr_noproc(transp);
      return;
  }
  memset((char *)&argument, 0, sizeof(argument));
  if (!svc_getargs(transp, _xdr_argument, (caddr_t)&argument)) {
    svcerr_decode(transp);
    return;
  }
  result = (*local)((char *)&argument, rqstp);
  if (result != NULL && !svc_sendreply(transp, _xdr_result, result)) {
    svcerr_systemerr(transp);
  }
  if (!svc_freeargs(transp, _xdr_argument, (caddr_t) &argument)) {
    fprintf(stderr, "unable to free arguments\n");
    exit(1);
  }
  return;
}

void dispatch_nfs(struct svc_req *rqstp, register SVCXPRT *transp)
{
  union {
    nfs_fh nfsproc_getattr_2_arg;
    sattrargs nfsproc_setattr_2_arg;
    diropargs nfsproc_lookup_2_arg;
    nfs_fh nfsproc_readlink_2_arg;
    readargs nfsproc_read_2_arg;
    writeargs nfsproc_write_2_arg;
    createargs nfsproc_create_2_arg;
    diropargs nfsproc_remove_2_arg;
    renameargs nfsproc_rename_2_arg;
    linkargs nfsproc_link_2_arg;
    symlinkargs nfsproc_symlink_2_arg;
    createargs nfsproc_mkdir_2_arg;
    diropargs nfsproc_rmdir_2_arg;
    readdirargs nfsproc_readdir_2_arg;
    nfs_fh nfsproc_statfs_2_arg;
  } argument;
  char *result;
  xdrproc_t _xdr_argument, _xdr_result;
  char *(*local)(char *, struct svc_req *);

  switch (rqstp->rq_proc) {
    case NFSPROC_NULL:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_null_2_svc;
      break;

    case NFSPROC_GETATTR:
      _xdr_argument = (xdrproc_t) xdr_nfs_fh;
      _xdr_result = (xdrproc_t) xdr_attrstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_getattr_2_svc;
      break;

    case NFSPROC_SETATTR:
      _xdr_argument = (xdrproc_t) xdr_sattrargs;
      _xdr_result = (xdrproc_t) xdr_attrstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_setattr_2_svc;
      break;

    case NFSPROC_ROOT:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_root_2_svc;
      break;

    case NFSPROC_LOOKUP:
      _xdr_argument = (xdrproc_t) xdr_diropargs;
      _xdr_result = (xdrproc_t) xdr_diropres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_lookup_2_svc;
      break;

    case NFSPROC_READLINK:
      _xdr_argument = (xdrproc_t) xdr_nfs_fh;
      _xdr_result = (xdrproc_t) xdr_readlinkres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_readlink_2_svc;
      break;

    case NFSPROC_READ:
      _xdr_argument = (xdrproc_t) xdr_readargs;
      _xdr_result = (xdrproc_t) xdr_readres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_read_2_svc;
      break;

    case NFSPROC_WRITECACHE:
      _xdr_argument = (xdrproc_t) xdr_void;
      _xdr_result = (xdrproc_t) xdr_void;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_writecache_2_svc;
      break;

    case NFSPROC_WRITE:
      _xdr_argument = (xdrproc_t) xdr_writeargs;
      _xdr_result = (xdrproc_t) xdr_attrstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_write_2_svc;
      break;

    case NFSPROC_CREATE:
      _xdr_argument = (xdrproc_t) xdr_createargs;
      _xdr_result = (xdrproc_t) xdr_diropres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_create_2_svc;
      break;

    case NFSPROC_REMOVE:
      _xdr_argument = (xdrproc_t) xdr_diropargs;
      _xdr_result = (xdrproc_t) xdr_nfsstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_remove_2_svc;
      break;

    case NFSPROC_RENAME:
      _xdr_argument = (xdrproc_t) xdr_renameargs;
      _xdr_result = (xdrproc_t) xdr_nfsstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_rename_2_svc;
      break;

    case NFSPROC_LINK:
      _xdr_argument = (xdrproc_t) xdr_linkargs;
      _xdr_result = (xdrproc_t) xdr_nfsstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_link_2_svc;
      break;

    case NFSPROC_SYMLINK:
      _xdr_argument = (xdrproc_t) xdr_symlinkargs;
      _xdr_result = (xdrproc_t) xdr_nfsstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_symlink_2_svc;
      break;

    case NFSPROC_MKDIR:
      _xdr_argument = (xdrproc_t) xdr_createargs;
      _xdr_result = (xdrproc_t) xdr_diropres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_mkdir_2_svc;
      break;

    case NFSPROC_RMDIR:
      _xdr_argument = (xdrproc_t) xdr_diropargs;
      _xdr_result = (xdrproc_t) xdr_nfsstat;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_rmdir_2_svc;
      break;

    case NFSPROC_READDIR:
      _xdr_argument = (xdrproc_t) xdr_readdirargs;
      _xdr_result = (xdrproc_t) xdr_readdirres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_readdir_2_svc;
      break;

    case NFSPROC_STATFS:
      _xdr_argument = (xdrproc_t) xdr_nfs_fh;
      _xdr_result = (xdrproc_t) xdr_statfsres;
      local = (char *(*)(char *, struct svc_req *)) nfsproc_statfs_2_svc;
      break;

    default:
      svcerr_noproc (transp);
      return;
  }
  memset((char *)&argument, 0, sizeof (argument));
  if (!svc_getargs(transp, _xdr_argument, (caddr_t)&argument)) {
    svcerr_decode(transp);
    return;
  }

  result = (*local)((char *)&argument, rqstp);
  if (result != NULL && !svc_sendreply(transp, _xdr_result, result)) {
    svcerr_systemerr(transp);
  }
  if (!svc_freeargs(transp, _xdr_argument, (caddr_t)&argument)) {
    fprintf(stderr, "unable to free arguments\n");
    exit(1);
  }
  return;
}

void handle_c(SV *readfds) {
  int  readfds_len;
  char *readfds_s;

  readfds_s = SvPV(readfds, readfds_len);

  if (readfds_len != sizeof(svc_fdset)) {
    fprintf(stderr, "handle_c: readfds is wrong length (%d)\n", readfds_len);
    return;
  }

  svc_getreqset((fd_set *)readfds_s);
}

int init_mount(void) {
  register SVCXPRT *transp;

  pmap_unset(KATIE_MOUNTPROG, KATIE_MOUNTVERS);
  pmap_unset(KATIE_MOUNTPROG, KATIE_MOUNTVERS_POSIX);

  transp = svcudp_create(RPC_ANYSOCK);
  if (transp == NULL) {
    fprintf(stderr, "cannot create udp service.\n");
    return 0;
  }
  if (!svc_register(transp, KATIE_MOUNTPROG, KATIE_MOUNTVERS,
                    dispatch_mount, IPPROTO_UDP)) {
    fprintf(stderr, "unable to register (KATIE_MOUNTPROG, 1, udp).\n");
    return 0;
  }
  if (!svc_register(transp, KATIE_MOUNTPROG, KATIE_MOUNTVERS_POSIX,
                    dispatch_mount, IPPROTO_UDP)) {
    fprintf(stderr, "unable to register (KATIE_MOUNTPROG, 2, udp).\n");
    return 0;
  }

  return 1;
}

int init_nfs(void) {
  register SVCXPRT *transp;

  pmap_unset(KATIE_PROGRAM, KATIE_VERSION);

  transp = svcudp_create(RPC_ANYSOCK);
  if (transp == NULL) {
    fprintf(stderr, "cannot create udp service.\n");
    return 0;
  }
  if (!svc_register(transp, KATIE_PROGRAM, KATIE_VERSION,
                    dispatch_nfs, IPPROTO_UDP)) {
    fprintf(stderr, "unable to register (KATIE_PROGRAM, udp).\n");
    return 0;
  }
}

SV *init(
    SV *mnt_handler_in,
    SV *getattr_handler_in,
    SV *setattr_handler_in,
    SV *lookup_handler_in,
    SV *read_handler_in,
    SV *write_handler_in,
    SV *create_handler_in,
    SV *remove_handler_in,
    SV *rename_handler_in,
    SV *link_handler_in,
    SV *mkdir_handler_in,
    SV *rmdir_handler_in,
    SV *readdir_handler_in,
    SV *statfs_handler_in
) {
  CLIENT *clnt;
  Inline_Stack_Vars;

  if (!init_mount() || !init_nfs()) {
    return &PL_sv_undef;
  }

  PERL_HASH(uid_hash, "uid", 3);
  PERL_HASH(gid_hash, "gid", 3);
  PERL_HASH(groups_hash, "groups", 6);

  mnt_handler = newSVsv(mnt_handler_in);
  getattr_handler = newSVsv(getattr_handler_in);
  setattr_handler = newSVsv(setattr_handler_in);
  lookup_handler = newSVsv(lookup_handler_in);
  read_handler = newSVsv(read_handler_in);
  write_handler = newSVsv(write_handler_in);
  create_handler = newSVsv(create_handler_in);
  remove_handler = newSVsv(remove_handler_in);
  rename_handler = newSVsv(rename_handler_in);
  link_handler = newSVsv(link_handler_in);
  mkdir_handler = newSVsv(mkdir_handler_in);
  rmdir_handler = newSVsv(rmdir_handler_in);
  readdir_handler = newSVsv(readdir_handler_in);
  statfs_handler = newSVsv(statfs_handler_in);

  return newSVpv((char *)&svc_fdset, sizeof(svc_fdset));
}
