# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::Constants;

use strict;

require Exporter;
@Katie::Constants::ISA = qw(Exporter);
@Katie::Constants::EXPORT = qw(
  KATIE_SERVER_INFO_FN
  KATIE_SERVER_INFO_INODE
  PREFIX_DIR
  DATA_DIR
  RUN_DIR
  MAN_DIR
  DB_CONF_FN
  DEFAULT_CONFIG_SPEC_FN
  KATIE_HELP_FN

  BUILT_IN_DEFAULT_CS
  MAX_SMALL_FILE_SIZE

  ROOT_ELEMENT_ID
  MAIN_BRANCH_ID

  iELEM
  iATAT
  iELBR
  iREVI
  iVUPR
  iCOUT
  iBRTY

  dVIEW
  dATAT
  dELBR
  dREVI
  dVUPR
  dCOUT

  eDIR
  eREG
);

# name of the file in which the katie server name, etc. appear in
use constant KATIE_SERVER_INFO_FN => ".katie-server-info";

# inode number of KATIE_SERVER_INFO_FN
use constant KATIE_SERVER_INFO_INODE => 0;

# directory that Katie is installed in
use constant PREFIX_DIR => '%PREFIX%';

# directory for run-time documentation
use constant DOC_DIR => '%DOCDIR%';

# directory for data files (eg. VOB, view store)
use constant DATA_DIR => '%DATADIR%';

# directory for runtime data (eg. ktab)
use constant RUN_DIR => '%RUNDIR%';

# directory for man pages
use constant MAN_DIR => '%MANDIR%';

# name of the file with the database connection options
use constant DB_CONF_FN => PREFIX_DIR . "/share/katie/db.conf";

# name of the file containing the default config spec
use constant DEFAULT_CONFIG_SPEC_FN =>
                 PREFIX_DIR . "/share/katie/default_config_spec";

# name of the katie help file
use constant KATIE_HELP_FN => DOC_DIR . "/katie.help";

# built in default config spec used if the default is not usuable
use constant BUILT_IN_DEFAULT_CS => "
  element . CHECKEDOUT
  element . /main/LATEST
";

# maximum size of a config spec or comments file (for sanity checking)
use constant MAX_SMALL_FILE_SIZE => 65536;

# element_id of the root directory
use constant ROOT_ELEMENT_ID => 2;

# branch_id of 'main'
use constant MAIN_BRANCH_ID => 1;

# get_inode_type() returns these codes
# (should be get_object_type())
use constant iELEM => 1; # object is an element as seen in a view
use constant iATAT => 2; # object is an element's @@ directory
use constant iELBR => 3; # object is an element's branch
use constant iREVI => 4; # object is an element's revision
use constant iVUPR => 5; # object is view-private
use constant iCOUT => 6; # object is checked-out
use constant iBRTY => 6; # object is a branch type

# get_dir_type() returns these codes (and assumes that these have the
# same numbers as the above $i* variables);
use constant dVIEW => 1; # directory is an element seen through a view
use constant dATAT => 2; # directory is an element@@ directory
use constant dELBR => 3; # directory is an element@@/branch directory
use constant dREVI => 4; # directory is a version of a directory element
                         # (directory@@/branch/rev)
use constant dVUPR => 5; # directory is a view-private directory
use constant dCOUT => 6; # directory is a checkedout directory

# these must match up with the entries in the 'type' table
use constant eDIR => 1; # element is a directory
use constant eREG => 2; # element is a regular file

1;
