# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::NFSerrno;

use strict;

require Exporter;
@Katie::NFSerrno::ISA = qw(Exporter);
@Katie::NFSerrno::EXPORT = qw(
  nfsstrerr
  NFS_OK
  NFSERR_PERM
  NFSERR_NOENT
  NFSERR_IO
  NFSERR_NXIO
  NFSERR_ACCES
  NFSERR_EXIST
  NFSERR_NODEV
  NFSERR_NOTDIR
  NFSERR_ISDIR
  NFSERR_INVAL
  NFSERR_FBIG
  NFSERR_NOSPC
  NFSERR_ROFS
  NFSERR_NAMETOOLONG
  NFSERR_NOTEMPTY
  NFSERR_DQUOT
  NFSERR_STALE
  NFSERR_WFLUSH
);

use constant NFS_OK             => 0;
use constant NFSERR_PERM        => 1;
use constant NFSERR_NOENT       => 2;
use constant NFSERR_IO          => 5;
use constant NFSERR_NXIO        => 6;
use constant NFSERR_ACCES       => 13;
use constant NFSERR_EXIST       => 17;
use constant NFSERR_NODEV       => 19;
use constant NFSERR_NOTDIR      => 20;
use constant NFSERR_ISDIR       => 21;
use constant NFSERR_INVAL       => 21;
use constant NFSERR_FBIG        => 27;
use constant NFSERR_NOSPC       => 28;
use constant NFSERR_ROFS        => 30;
use constant NFSERR_NAMETOOLONG => 63;
use constant NFSERR_NOTEMPTY    => 66;
use constant NFSERR_DQUOT       => 69;
use constant NFSERR_STALE       => 70;
use constant NFSERR_WFLUSH      => 99;

my(%errors) = (
  0  => "No error",
  1  => "Permission denied",
  2  => "No such file or directory",
  5  => "I/O error",
  6  => "No such device or address",
  13 => "Permission denied",
  17 => "File exists",
  19 => "No such device",
  20 => "Not a directory",
  21 => "Is a directory",
  22 => "Invalid argument",
  27 => "File too large",
  28 => "No space left on device",
  30 => "Read-only file system",
  63 => "File name too long",
  66 => "Directory not empty",
  69 => "Disc quota exceeded",
  70 => "Stale NFS file handle",
  99 => "Write cache flushed",
);

sub nfsstrerr {
  my($errno) = @_;

  if ($errors{$errno}) {
    return $errors{$errno};
  }
  return $errors{0};
}

1;
